@@ -33,11 +33,8 @@ The Free Software Foundation is independent of Sun Microsystems, Inc.  */
 
 static void set_constant_entry (CPool *, int, int, jword);
 static int find_tree_constant (CPool *, int, tree);
-static int find_class_or_string_constant (CPool *, int, tree);
 static int find_name_and_type_constant (CPool *, tree, tree);
 static tree get_tag_node (int);
-static tree build_constant_data_ref (void);
-static CPool *cpool_for_class (tree);
 
 /* Set the INDEX'th constant in CPOOL to have the given TAG and VALUE. */
 
@@ -133,7 +130,7 @@ find_utf8_constant (CPool *cpool, tree name)
   return find_tree_constant (cpool, CONSTANT_Utf8, name);
 }
 
-static int
+int
 find_class_or_string_constant (CPool *cpool, int tag, tree name)
 {
   jword j = find_utf8_constant (cpool, name);
@@ -321,6 +318,9 @@ get_tag_node (int tag)
 {
   /* A Cache for build_int_cst (CONSTANT_XXX, 0). */
 
+  if (tag >= 13)
+    return build_int_cst (NULL_TREE, tag);
+
   if (tag_nodes[tag] == NULL_TREE)
     tag_nodes[tag] = build_int_cst (NULL_TREE, tag);
   return tag_nodes[tag];
@@ -328,7 +328,7 @@ get_tag_node (int tag)
 
 /* Given a class, return its constant pool, creating one if necessary.  */
 
-static CPool *
+CPool *
 cpool_for_class (tree class)
 {
   CPool *cpool = TYPE_CPOOL (class);
@@ -422,14 +422,36 @@ alloc_class_constant (tree clas)
 
 /* Return the decl of the data array of the current constant pool. */
 
-static tree
-build_constant_data_ref (void)
+tree
+build_constant_data_ref (bool indirect)
 {
-  tree decl = TYPE_CPOOL_DATA_REF (output_class);
-
-  if (decl == NULL_TREE)
+  if (indirect)
     {
-      tree type;
+      tree d;
+      tree cpool_type = build_array_type (ptr_type_node, NULL_TREE);
+      tree decl = build_class_ref (output_class);
+      tree klass = build1 (INDIRECT_REF, TREE_TYPE (TREE_TYPE (decl)),
+			   decl);
+      tree constants = build3 (COMPONENT_REF, 
+			       TREE_TYPE (constants_field_decl_node), klass,
+			       constants_field_decl_node,
+			       NULL_TREE);
+      tree data = build3 (COMPONENT_REF, 
+			  TREE_TYPE (constants_data_field_decl_node), 
+			  constants,
+			  constants_data_field_decl_node,
+			  NULL_TREE);
+
+      TREE_THIS_NOTRAP (klass) = 1;
+      data = fold_convert (build_pointer_type (cpool_type), data);
+      d = build1 (INDIRECT_REF, cpool_type, data);
+      TREE_INVARIANT (d) = 1;
+
+      return d;
+    }
+  else
+    {
+      tree type, decl;
       tree decl_name = mangled_classname ("_CD_", output_class);
 
       /* Build a type with unspecified bounds.  The will make sure
@@ -444,10 +466,9 @@ build_constant_data_ref (void)
 
       decl = build_decl (VAR_DECL, decl_name, type);
       TREE_STATIC (decl) = 1;
-      TYPE_CPOOL_DATA_REF (output_class) = decl;
-    }
 
-  return decl;
+      return decl;
+    }
 }
 
 /* Get the pointer value at the INDEX'th element of the constant pool. */
@@ -455,27 +476,13 @@ build_constant_data_ref (void)
 tree
 build_ref_from_constant_pool (int index)
 {
-  tree d = build_constant_data_ref ();
-  tree i = build_int_cst (NULL_TREE, index);
-  if (flag_indirect_classes)
-    {
-      tree decl = build_class_ref (output_class);
-      tree klass = build1 (INDIRECT_REF, TREE_TYPE (TREE_TYPE (decl)),
-			   decl);
-      tree constants = build3 (COMPONENT_REF, 
-			       TREE_TYPE (constants_field_decl_node), klass,
-			       constants_field_decl_node,
-			       NULL_TREE);
-      tree data = build3 (COMPONENT_REF, 
-			  TREE_TYPE (constants_data_field_decl_node), 
-			  constants,
-			  constants_data_field_decl_node,
-			  NULL_TREE);
-      data = fold_convert (build_pointer_type (TREE_TYPE (d)), data);
-      d = build1 (INDIRECT_REF, TREE_TYPE (d), data);
-      /* FIXME: These should be cached.  */
-      TREE_INVARIANT (d) = 1;
-    }
+  tree i;
+  tree d = TYPE_CPOOL_DATA_REF (output_class);
+
+  if (d == NULL_TREE)
+    d = build_constant_data_ref (flag_indirect_classes);
+
+  i = build_int_cst (NULL_TREE, index);
   d = build4 (ARRAY_REF, TREE_TYPE (TREE_TYPE (d)), d, i,
 		 NULL_TREE, NULL_TREE);
   TREE_INVARIANT (d) = 1;
@@ -494,25 +501,33 @@ build_constants_constructor (void)
   tree tags_list = NULL_TREE;
   tree data_list = NULL_TREE;
   int i;
+
   for (i = outgoing_cpool->count;  --i > 0; )
-    switch (outgoing_cpool->tags[i])
+    switch (outgoing_cpool->tags[i] & ~CONSTANT_LazyFlag)
       {
+      case CONSTANT_None:  /* The second half of a Double or Long on a
+			      32-bit target.  */
       case CONSTANT_Fieldref:
       case CONSTANT_NameAndType:
+      case CONSTANT_Float:
+      case CONSTANT_Integer:
+      case CONSTANT_Double:
+      case CONSTANT_Long:
+      case CONSTANT_Methodref:
+      case CONSTANT_InterfaceMethodref:
 	{
 	  unsigned HOST_WIDE_INT temp = outgoing_cpool->data[i].w;
 
-	  /* Make sure that on a 64-bit big-endian machine this
-	     32-bit jint appears in the first word.  
+	  /* Make sure that on a big-endian machine with 64-bit
+	     pointers this 32-bit jint appears in the first word.
 	     FIXME: This is a kludge.  The field we're initializing is
 	     not a scalar but a union, and that's how we should
 	     represent it in the compiler.  We should fix this.  */
-	  if (BYTES_BIG_ENDIAN && BITS_PER_WORD > 32)
-	    temp <<= BITS_PER_WORD - 32;
+	  if (BYTES_BIG_ENDIAN && POINTER_SIZE > 32)
+	    temp <<= POINTER_SIZE - 32;
 
 	  tags_list
-	    = tree_cons (NULL_TREE, 
-			 build_int_cst (NULL_TREE, outgoing_cpool->tags[i]),
+	    = tree_cons (NULL_TREE, get_tag_node (outgoing_cpool->tags[i]),
 			 tags_list);
 	  data_list
 	    = tree_cons (NULL_TREE, 
@@ -521,7 +536,11 @@ build_constants_constructor (void)
 			 data_list);
 	}
 	break;
-      default:
+
+      case CONSTANT_Class:
+      case CONSTANT_String:
+      case CONSTANT_Unicode:
+      case CONSTANT_Utf8:
 	tags_list
 	  = tree_cons (NULL_TREE, get_tag_node (outgoing_cpool->tags[i]),
 		       tags_list);
@@ -529,6 +548,9 @@ build_constants_constructor (void)
 	  = tree_cons (NULL_TREE, build_utf8_ref (outgoing_cpool->data[i].t),
 		       data_list);
 	break;
+
+      default:
+	gcc_assert (false);
       }
   if (outgoing_cpool->count > 0)
     {
@@ -540,7 +562,7 @@ build_constants_constructor (void)
       tags_list = tree_cons (NULL_TREE, get_tag_node (0), tags_list);
       data_list = tree_cons (NULL_TREE, null_pointer_node, data_list);
   
-      data_decl = build_constant_data_ref ();
+      data_decl = build_constant_data_ref (false);
       TREE_TYPE (data_decl) = build_array_type (ptr_type_node, index_type);
       DECL_INITIAL (data_decl) = build_constructor_from_list
 				  (TREE_TYPE (data_decl), data_list);
