@@ -7,7 +7,7 @@
    PR:		none.
    Originator:	<andreast@gcc.gnu.org> 20030914	 */
 
-/* { dg-do run { xfail mips64*-*-* arm*-*-* strongarm*-*-* xscale*-*-* } } */
+/* { dg-do run } */
 #include "ffitest.h"
 
 typedef struct cls_struct_9byte {
@@ -68,7 +68,7 @@ int main (void)
   struct cls_struct_9byte res_dbl;
 
   cls_struct_fields[0] = &ffi_type_double;
-  cls_struct_fields[1] = &ffi_type_uint32;
+  cls_struct_fields[1] = &ffi_type_sint;
   cls_struct_fields[2] = NULL;
 
   dbl_arg_types[0] = &cls_struct_type;
